/********************************************************************
 * FILE: gomo.c
 * AUTHOR: Fabian Buske
 * CREATE DATE: 18/06/2008
 * PROJECT: MEME suite
 * COPYRIGHT: 2008, UQ
 *
 * gomo is an implementation of the algorithm described in
 * "Associating transcription factor binding site motifs with target
 * Go terms and target genes"
 * authors: Mikael Boden and Timothy L. Bailey
 * published: Nucl. Acids Res (2008)
 *
 ********************************************************************/

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <ctype.h>
#include <time.h>
#include "matrix.h"
#include "alphabet.h"
#include "cisml.h"
#include "fasta-io.h"
#include "meme-io.h"
#include "string-list.h"
#include "hash_table.h"
#include "simple-getopt.h"
#include "read_csv.h"
#include "ranksum_test.h"
#include "io.h"
#include "read_tamo.h"
#include "dir.h"
#include "qvalue.h"
#include "xml-util.h"
#include "config.h"
#include "projrel.h"

#include <libxslt/xslt.h>
#include <libxslt/xsltInternals.h>
#include <libxslt/transform.h>
#include <libxslt/xsltutils.h>

#define min(a,b)      (a<b)?a:b
#define max(a,b)      (a<b)?b:a
#define concat(a, b)  a ## b

char* program_name = "GOMO";
VERBOSE_T verbosity = NORMAL_VERBOSE;
const int GFF_FORMAT=0;   // GFF file format
const int CISML_FORMAT=1; // cisML file format
#define HASH_BINS 1000 	/* Number of bins used for hashing. */
#define SHOW_EVALUE 10 	/* E-value threshold */
#define MCHUNK 10 			/* size of memory chunks to allocate for lines */

static BOOLEAN clobber = FALSE; /* default is not to overwrite existing files */
static char *default_output_dirname = "gomo_out";  /* default name of output */

/* gomo result */
typedef struct gomo_result GOMO_RESULT_T;
struct gomo_result {
  char* goterm;	/* go term */
  double score;	/* score for go term */
  int index;
};

/*************************************************************************
 * gomo_dts()
 *************************************************************************/
static char *gomo_dts = "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n"
	"<?xml-stylesheet type=\"text/xsl\" href=\"gomo.xsl\"?>\n"
	"<!-- Document definition -->\n"
	"<!DOCTYPE gomo[\n"
	"<!ELEMENT gomo (program-name, model, species )>\n"
	"<!ATTLIST gomo xmlns:xsi CDATA #IMPLIED\n"
	"	version CDATA #REQUIRED\n"
	"	release CDATA #REQUIRED\n"
	">\n"
	"<!ELEMENT program-name (#PCDATA)>\n"
	"<!ELEMENT model (command_line, pvalue)>\n"
	"<!ELEMENT command_line (#PCDATA)>\n"
	"<!ELEMENT significance_threshold (#PCDATA)>\n"
	"<!ELEMENT species (input-files, (pattern+) )>\n"
	"<!ATTLIST species name CDATA #REQUIRED\n"
	"   id	CDATA #REQUIRED\n"
	">\n"
	"<!ELEMENT input-files (scoring-file, godb-file)>\n"
	"<!ELEMENT scoring-file (#PCDATA)>\n"
	"<!ELEMENT godb-file (#PCDATA)>\n"
	"<!ELEMENT pattern (goterm*)>\n"
	"<!ATTLIST pattern accession CDATA #REQUIRED\n"
	"	name CDATA #REQUIRED\n"
	">\n"
	"<!ELEMENT goterm (#PCDATA)>\n"
	"<!ATTLIST goterm id CDATA #REQUIRED\n"
	"	description CDATA #REQUIRED\n"
	"	E-value CDATA #REQUIRED\n"
	"	qvalue CDATA #REQUIRED\n"
	">\n"
	"]>\n";

/*************************************************************************
 * add_to_hash()
 *
 * Add one sequence go-term pair to the hash
 *************************************************************************/
inline void add_to_hash(char* go_term, HASH_TABLE* go_ht, char* seq_id){
	assert(go_term != NULL);
	assert(go_ht != NULL);
	assert(seq_id != NULL);

	int dummy; // dummy: required by hash_lookup interface.
	HASH_TABLE_ENTRY* he = NULL;
	he = hash_lookup_str(go_term, *go_ht, &dummy);
	if (he == NULL) {
		// Record current go term
		BOOLEAN b = hash_insert_str(go_term, *go_ht);
		if (b==FALSE)
			die("Hashing of go term failed!");
		else
			he = hash_lookup_str(go_term, *go_ht, &dummy);
	}
	if (hash_get_entry_value(he) == NULL){
		hash_set_entry_value(new_string_list(),he);
	}
	add_string(seq_id, (STRING_LIST_T*)(hash_get_entry_value(he)));
}

/*************************************************************************
 * add_to_description_hash()
 *
 * Adds one go term - description pair to the hash
 *************************************************************************/
inline void add_to_description_hash(char* go_term, HASH_TABLE* godesc_ht,
		char* go_desc){

	assert(go_term != NULL);
	assert(go_desc != NULL);
	assert(godesc_ht != NULL);

	int dummy; // dummy: required by hash_lookup interface.
	HASH_TABLE_ENTRY* he = NULL;
	he = hash_lookup_str(go_term, *godesc_ht, &dummy);
	if (he == NULL) {
		// add go term to go-description hash
		if (hash_insert_str(go_term, *godesc_ht)){
			he = hash_lookup_str(go_term, *godesc_ht, &dummy);
			// add description as the value
			hash_set_entry_value(go_desc,he);
		} else
			die("ERROR: Hashing of go description failed.");
	}
}

/*************************************************************************
 * get_unique_go_terms()
 *
 * Reads all go files for a gomo run and returns a unique list of all the
 * go_terms contained in the files
 *************************************************************************/
void get_unique_go_terms(
		STRING_LIST_T* gofiles,	// all go files to scan IN
		STRING_LIST_T* go_terms // STRING_LIST holding the set of unique go terms OUT
){
	assert(go_terms != NULL);
	assert(get_num_strings(go_terms) == 0);
	assert(gofiles != NULL);

	if (verbosity >= HIGH_VERBOSE){
		fprintf(stderr,"Read in %d go files in total.\n", get_num_strings(gofiles));
	}

	int c,i,j;
	for (c=0; c<get_num_strings(gofiles);++c){
		int length = 0;
		FILE *inputFilePtr;            	/* Pointer to input file */
		STRING_LIST_T **content = NULL;	/* content of input file */

		if (verbosity >= HIGH_VERBOSE){
			fprintf(stderr,"Read file %s...\n",get_nth_string(c,gofiles));
		}

		inputFilePtr = fopen(get_nth_string(c,gofiles), "r");  	/* Open in TEXT mode */
		read_csv(inputFilePtr,'#','\t',&content,&length); 		/* Get csv content*/
		fclose(inputFilePtr);

		for (i=0;i<length;++i){ /* read in each and every line */
			if (get_num_strings(content[i])>0){
				char* go_term = get_nth_string(0,content[i]);
				if (!have_string(go_term,go_terms)){
					add_string(go_term, go_terms);
				}
			}
			// free content
			free_string_list(content[i]);
		}
		myfree(content);
	}

	if (verbosity >= NORMAL_VERBOSE){
		fprintf(stderr, "Total number of unique go terms %d\n", get_num_strings(go_terms));
	}
}

/*************************************************************************
 * read_go_file()
 *
 * Reads in go term to sequence file
 *************************************************************************/
void read_go_file(char* filename, STRING_LIST_T* go_terms, HASH_TABLE* go_ht, HASH_TABLE* godesc_ht){
	FILE *inputFilePtr;            	/* Pointer to input file */
	STRING_LIST_T **content = NULL;

	assert(filename != NULL);
	assert(go_terms != NULL);
	assert(go_ht != NULL);
	assert(godesc_ht != NULL);

	int i,j,hash_value,length = 0;
	inputFilePtr = fopen(filename, "r");  	/* Open in TEXT mode */
	read_csv(inputFilePtr,'#','\t',&content,&length); /* Get csv content*/
	fclose(inputFilePtr);

	if (verbosity >= NORMAL_VERBOSE){
		fprintf(stderr, "Read in %d lines from %s\n", length,filename);
	}

	for (i=0;i<length;++i){ /* read in each and every line */
		if (get_num_strings(content[i])<=2){
			if (verbosity >= NORMAL_VERBOSE){
				fprintf(stderr, "Go-term does not exists or lacks assigned sequences. Skip this entry!\n");
			}
		} else {
			char* go_term = get_nth_string(0,content[i]);
			// link sequence ids to go term via hash
			for (j=2;j<get_num_strings(content[i]);++j){
				char* seq_id = get_nth_string(j,content[i]);
				add_to_hash(go_term, go_ht, seq_id);
			}
			/* add go-term to key list if not included yet */
			if (!have_string(go_term,go_terms)){
				add_string(go_term, go_terms);
			}
			/* add description - since content will be freed copy description first */
			char* go_desc = NULL;
			Resize(go_desc, strlen(get_nth_string(1,content[i]))+1, char);
			go_desc = strcpy(go_desc,get_nth_string(1,content[i]));
			add_to_description_hash(go_term, godesc_ht, go_desc);
		}
	}
	// free content
	for (i=0;i<length;++i){
		free_string_list(content[i]);
	}
	myfree(content);
}

/*************************************************************************
 * remove_empty_goterms()
 *
 * Remove the go terms from the list of terms that lack any associated
 * sequence
 *************************************************************************/
void remove_empty_goterms(STRING_LIST_T* go_terms, HASH_TABLE* go_ht){
	int i, dummy;
	HASH_TABLE_ENTRY* hex = NULL;
	STRING_LIST_T *remove = new_string_list();
	for (i=0;i<get_num_strings(go_terms);++i){
		char* go_term = get_nth_string(i,go_terms);
		hex = hash_lookup_str(go_term, *go_ht, &dummy);
		if (hex == NULL){
			add_string(go_term,remove);
		}
	}
	if (get_num_strings(remove)>0){
		remove_strings(remove, go_terms);
	}
	free_string_list(remove);
}

/*************************************************************************
 * gomo_result_cmp()
 *
 * Compares two gomo results with respect to its scores
 *************************************************************************/
inline int gomo_result_cmp(const void *v1, const void *v2)
{
  assert(v1 != NULL);
  assert(v2 != NULL);
  GOMO_RESULT_T s1 = *(GOMO_RESULT_T *)v1;
  GOMO_RESULT_T s2 = *(GOMO_RESULT_T *)v2;
  double diff = s1.score - s2.score;
  if (diff > 0.0){
	  return 1;
  } else if (diff < 0.0){
	  return -1;
  } else {
	  return 0;
  }
}

/*************************************************************************
 * sequence_cmp()
 *
 * Compares two scanned sequences with regard to their scores
 *************************************************************************/
inline int sequence_cmp(const void *v1, const void *v2)
{
  assert(v1 != NULL);
  assert(v2 != NULL);
  SCANNED_SEQUENCE_T* s1 = *(SCANNED_SEQUENCE_T **)v1;
  SCANNED_SEQUENCE_T* s2 = *(SCANNED_SEQUENCE_T **)v2;
  double diff = get_scanned_sequence_score(s1)-get_scanned_sequence_score(s2);
  if (diff < 0.0){
	  return 1;
  } else if (diff > 0.0){
	  return -1;
  } else {
	  return 0;
  }
}

/*************************************************************************
 * sort_cisml_sequences()
 *
 * Sorts the scanned sequences contained in the cisml data structure
 * regarding their p-value and grouped by the pattern
 *************************************************************************/
void sort_scanned_sequences(PATTERN_T *pattern){
  assert(pattern != NULL);
  SCANNED_SEQUENCE_T ** seq_arr = get_pattern_scanned_sequences(pattern);
  int size = get_pattern_num_scanned_sequences(pattern);
  qsort(seq_arr,size,sizeof(seq_arr[0]),sequence_cmp);

  if (verbosity >= HIGHER_VERBOSE){
	  int i;
	  fprintf(stderr,"Pattern: %s\nOrdered Sequences:\n",get_pattern_name(pattern));
	  for (i=0;i<size;++i){
		  SCANNED_SEQUENCE_T* s = get_pattern_scanned_sequences(pattern)[i];
		  fprintf(stderr,"%s : %f\n",get_scanned_sequence_accession(s),get_scanned_sequence_score(s));
	  }
  }
}

/*************************************************************************
 * hash_sequence_ranks()
 *
 * Hash the sequence ids to its ranks
 *************************************************************************/
void hash_sequence_ranks(HASH_TABLE* seqRanks_ht, PATTERN_T* pattern){
  assert(seqRanks_ht != NULL);
  assert(pattern != NULL);
  int seqs = get_pattern_num_scanned_sequences(pattern);

  // Hash mapping all sequences to its rank
  *seqRanks_ht = hash_create(seqs*4);
  int k,dummy;
  for (k=0;k<seqs;++k){
	  SCANNED_SEQUENCE_T *seq = get_pattern_scanned_sequences(pattern)[k];
	  char* seqName = get_scanned_sequence_accession(seq);
	  HASH_TABLE_ENTRY *he = hash_lookup_str(seqName,*seqRanks_ht,&dummy);
	  if (he == NULL){
		  BOOLEAN b = hash_insert_str(seqName, *seqRanks_ht);
		  he = hash_lookup_str(seqName, *seqRanks_ht, &dummy);
		  if (b==FALSE)
			  die("Hashing of sequences to its rank did not work out! bad luck!");

		  if (hash_get_entry_value(he) == NULL){
			  int* value = mm_malloc(sizeof(int));
			  *value = k;
			  hash_set_entry_value(value,he);
		  }
	  } else {
		  die("Error! Key already exists in the hash table: %s\n",seqName);
	  }
  }
}


/*************************************************************************
 * add_to_motifhash()
 *
 * adds a new motif to the motif hashtable if not contained already
 *************************************************************************/
void add_to_motifht(
		HASH_TABLE *motif_ht,
		STRING_LIST_T* motifs,
		STRING_LIST_T* all_go_terms,
		char* accession,
		int num_organisms
){
	int dummy,j,k;
	char* g;
	double* organisms;
	HASH_TABLE_ENTRY* he;
	HASH_TABLE tmp_ht;

	// add motif if not in motif hash yet
	he = hash_lookup_str(accession,*motif_ht, &dummy);
	if (he == NULL) {
		  // add motif to keyset
		  add_string(accession, motifs);
		  // add motif to hashtable
		  hash_insert_str(accession, *motif_ht);
		  // get corresponding hash entry
		  he = hash_lookup_str(accession, *motif_ht, &dummy);
		  // add go term hashtable as value
		  tmp_ht = hash_create(get_num_strings(all_go_terms)*2);
		  hash_set_entry_value(tmp_ht,he);
		  // fill goterm hashtable with all go_terms as key and
		  // assign an array of size = number of organisms
		  for (j=0;j<get_num_strings(all_go_terms);++j){
			  g = get_nth_string(j,all_go_terms);
			  hash_insert_str(g, tmp_ht);
			  he = hash_lookup_str(g, tmp_ht, &dummy);
			  organisms = mm_malloc(sizeof(double)*num_organisms);
			  for (k=0;k<num_organisms;++k)
				  organisms[k] = -1.0; // flag as not available / set yet
			  hash_set_entry_value(organisms,he);
		  }
	}
}

/*************************************************************************
 * run_gomo_on_organism()
 *
 *
 *************************************************************************/
void run_gomo_on_organism(
		CISML_T* cisml,			// CISML containing scored sequences and patterns IN
		PATTERN_T *pattern,		// pattern of interest IN
		HASH_TABLE go_ht,		// GO-term 2 seq id hashtable IN
		HASH_TABLE godesc_ht,	// GO-term 2 GO description hashtable IN
		STRING_LIST_T* organism_go_terms,	// all GO terms specific to this organism IN
		STRING_LIST_T* all_go_terms, // all unique Go terms across all organisms IN
		double e_threshold,		// threshold for E-value IN
		int organism,			// current organism of interest
		int num_organisms,		// all organisms
  		HASH_TABLE motif_ht,
		BOOLEAN_T status,		// FLAG whether to output the program status IN
		BOOLEAN_T text_only,	// FLAG whether to output text only IN
		FILE* xml_output,		// target for xml output IN
		FILE* text_output,		// target for txt output IN
		STRING_LIST_T* motifs 	// motifs processed for this organism OUT
){
  int i = 0;
  int j = 0;
  int k = 0;
  int dummy = 0;
  int s_length = 0;
  int class_a = 0;
  double p_value;
  HASH_TABLE seqRanks_ht = NULL;
  HASH_TABLE_ENTRY *he = NULL;
  char* go_term = NULL;
  int* selected = NULL;
  double* score_pred = NULL;
  double* scores = NULL;
  BOOLEAN_T* group = NULL;
  RSD_T* ranksum_dataset = NULL;
  RSR_T *ranksum_result = NULL;
  STRING_LIST_T* genelist = NULL;
  char* accession = NULL;		// sequence accession
  int seqs;				// number of sequences in a pattern
  ARRAY_T* pv = NULL;	// pvalue array
  double* v = NULL;		// p-values
  int orphants=0;		// GO terms with gene id but lacking any sequence for these

  if (status)
	  fprintf(stderr,"Process organism %d : 0.0 %% \r",organism+1);

  selected = NULL;
  s_length = 0;
  seqs = get_pattern_num_scanned_sequences(pattern);
  accession = get_pattern_accession(pattern);

  if (!text_only){
	  fprintf(xml_output,"<pattern accession=\"%s\" name=\"%s\">\n",accession,get_pattern_name(pattern));
  }

  // sort scanned sequences with regard to score
  sort_scanned_sequences(pattern);

  // add motif to motif hashtable if more than one organism is scanned
  if (num_organisms > 1){
	  add_to_motifht(&motif_ht, motifs, all_go_terms, accession, num_organisms);
  }

  // create hash that maps gene ids to its rank
  hash_sequence_ranks(&seqRanks_ht,pattern);

  // predicted score
  score_pred = mm_malloc(sizeof(double)*get_num_strings(organism_go_terms));
  // init with -1
  for (j=0;j<get_num_strings(organism_go_terms);++j){
	  score_pred[j] = -1.0;
  }
  // class of each sequence
  group = mm_malloc(sizeof(BOOLEAN_T)*seqs);
  // get the ordered scores list
  scores = mm_malloc(sizeof(double)*seqs);
  for (j=0;j<seqs;++j){
	  SCANNED_SEQUENCE_T *seq = get_pattern_scanned_sequences(pattern)[j];
	  scores[j] = get_scanned_sequence_score(seq);
  }

  // Copy the scores and the uninitialized classes (group) to
  // a special object for ranksum routines.  Calculate the
  // (adjusted) rank of each score.
  // The group of each score will be reassigned differently for
  // each GO term.
  ranksum_dataset = get_ranksum_dataset(scores, group, seqs);

  /* get list f sorted p-values and compute qvalues */
  Resize(pv,1,ARRAY_T);

  /* check each go-term */
  for (j=0;j<get_num_strings(organism_go_terms);++j){

	  // number of elements in class a
	  class_a = 0;

	  // get go-term (key)
	  go_term = get_nth_string(j,organism_go_terms);

	  // define all entries as not belonging to the GO-term by default
	  for (k=0;k<seqs;++k){
		  group[k] = FALSE;
	  }

	  // look up go-term associated gene ids
	  he = hash_lookup_str(go_term,go_ht,&dummy);
	  genelist = (STRING_LIST_T*)(hash_get_entry_value(he));

	  // for each gene id look up its rank
	  for (k=0;k<get_num_strings(genelist);++k){
		  // Get the geneid from the list and use it as a key
		  // to get the position (rank) of the gene in the sorted list.
		  HASH_TABLE_ENTRY *hlp = hash_lookup_str(get_nth_string(k,genelist),seqRanks_ht,&dummy);
		  if (hlp != NULL){
			group[*((int*)hash_get_entry_value(hlp))] = TRUE;
			++class_a;
		  }
	  }
	  // if not all genes are assigned to this go term but at least one is
	  // i.e. some information is found ... proceed
	  if (seqs > get_num_strings(genelist) && class_a > 0){

		  set_ranksum_groups(ranksum_dataset, group);
		  /* perform ranksum test on ordered dataset */
		  ranksum_result = run_ranksum_test_on_ordered_dataset(ranksum_dataset);

		  // returns p-value of left-tailed tests
		  p_value = RSR_get_p_left(ranksum_result);

		  // look up motif in motif-hashtable if more than one organism investigated
		  if (num_organisms > 1){
			  he = hash_lookup_str(accession, motif_ht, &dummy);
			  if (he != NULL){
				  // look up go term for motif
				  he = hash_lookup_str(go_term, (HASH_TABLE)he->value, &dummy);
				  if (he != NULL){
					  ((double*)he->value)[organism] = p_value;
				  } else {
					  die("Error: Could not lookup go term in hashtable. "
							  "Something during initialization must have failed.");
				  }
			  } else {
				  die("Error: Could not lookup motif in hashtable. "
						  "Something during initialization must have failed.");
			  }
		  }

		  // save p_value
		  score_pred[j] = p_value;
		  destroy_rsr(ranksum_result);
	  } else{
		  ++orphants;
		  if (verbosity >= HIGHER_VERBOSE){
			  fprintf(stderr,"Skip due to all or no GO term assigned to this gene:%d %d %s\n",get_num_strings(genelist), class_a,go_term);
		  }
	  }
  }

  if (verbosity >= HIGH_VERBOSE){
	  fprintf(stderr,"%d GO terms have assigned genes which lack any sequence only or for other reasons have no information to work with. These are skipped!\n",orphants);
  }
  int num_go_terms = get_num_strings(organism_go_terms)-orphants;

  myfree(scores);
  myfree(group);
  destroy_rsd(ranksum_dataset);

  /* order results with respect to p-value */
  GOMO_RESULT_T* res = NULL;

  Resize(v,num_go_terms,double);
  pv->items = v;
  pv->num_items = num_go_terms;


  /* correcting for multiple tests */
  int c=0;
  for (j=0;j<get_num_strings(organism_go_terms);++j){
	  if (score_pred[j]!= -1){
		  set_array_item(c++, score_pred[j], pv);
		  if (e_threshold <= 0.0 || score_pred[j]*num_go_terms<e_threshold){
			  /* resize array of pointers if necessary */
			  if ((s_length % MCHUNK) == 0){
				  Resize(selected, s_length+MCHUNK, int);
			  }
			  selected[s_length++] = j;
		  } else if (verbosity >= HIGHER_VERBOSE){
			  fprintf(stderr,"Not satisfying go term:%s %g\n", go_term, score_pred[j]*num_go_terms);
		  }
	  }
  }
  sort_array(FALSE,pv);
  compute_qvalues(FALSE,
		  TRUE,
		  NUM_BOOTSTRAPS,
		  NUM_BOOTSTRAP_SAMPLES,
		  NUM_LAMBDA,
		  MAX_LAMBDA,
		  num_go_terms,
		  pv);

  /* output results with respect to p-value */
  if (s_length > 0){
	  Resize(res,s_length,GOMO_RESULT_T);
	  for (j=0;j<s_length;++j){
		  char* go_term = get_nth_string(selected[j],organism_go_terms);
		  res[j].score = score_pred[selected[j]];
		  res[j].goterm = go_term;
		  res[j].index = selected[j];
	  }
	  qsort(res,s_length,sizeof(GOMO_RESULT_T),gomo_result_cmp);

	  /* output go-terms*/
	  for (j=0;j<s_length;++j){
		  char* go_desc = "";
		  he = hash_lookup_str(res[j].goterm,godesc_ht,&dummy);
		  if (he != NULL && hash_get_entry_value(he) != NULL){
			  go_desc = (char*)(hash_get_entry_value(he));
		  }
		  fprintf(text_output,"%s\t%s\t%.3e\t%.3e\t%s\n",get_pattern_accession(pattern),
				  get_nth_string(res[j].index,organism_go_terms),
				  res[j].score*num_go_terms,get_array_item(j,pv),go_desc);

		  if (!text_only){
			  fprintf(xml_output,"<goterm id=\"%s\" E-value=\"%.3e\" qvalue=\"%.3e\" description=\"%s\"/>\n",
					  res[j].goterm,res[j].score*num_go_terms,get_array_item(j,pv),go_desc);
		  }
	  }
  }

  fprintf(text_output,"# %d total go-terms out of %d satisfying E <= %g for motif %s \n",s_length,num_go_terms,e_threshold,get_pattern_accession(pattern));

  /**********
  * clean up
  ***********/

  // free sequence hashmap
  for (k=0;k<seqs;++k){
	  SCANNED_SEQUENCE_T *seq = get_pattern_scanned_sequences(pattern)[k];
	  char* seqName = get_scanned_sequence_accession(seq);
	  he = hash_lookup_str(seqName,seqRanks_ht,&dummy);
	  if (he != NULL && hash_get_entry_value(he) != NULL){
		  int* x = (int *)hash_get_entry_value(he);
		  myfree(x);
	  }
  }

  hash_destroy(seqRanks_ht);
  myfree(score_pred);
  myfree(selected);
  myfree(res);
  free_array(pv);

  if (!text_only){
	  fprintf(xml_output,"</pattern>\n");
  }

  if (status)
	  fprintf(stderr, "Process organism %d: %3.1f %% \r",organism+1,(i)*100.0/(float)get_cisml_num_patterns(cisml));
}

/*************************************************************************
 * start_output()
 *
 * initiate the output
 *************************************************************************/
void start_output(FILE *text_output,
		FILE *xml_output,
		BOOLEAN_T text_only,
		char* program_name,
		double e_threshold,
		int argc,
		char *argv[]
){
  char* command = NULL;
  int lc = 4;
  int alloc_size = 100;
  int i = 0;
  int j = 0;

  fprintf(text_output,"# GOMO %s\n",VERSION);
  fprintf(text_output,"# Motif accession number \tGO-ID\tE-value\tq-value\tGO description\n");

  const char *archive_date = ARCHIVE_DATE;
  int a = strlen(archive_date);

  if (!text_only) {
	  fprintf(xml_output,gomo_dts);
	  fprintf(xml_output,
	    "<gomo xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" version=\"%s\" release=\"%.*s\">\n",
	    VERSION,
	    a,
	    archive_date
	  );
	  fprintf(xml_output,"<program-name>%s</program-name>\n",program_name);

	  // create command line
	  Resize(command, alloc_size, char);
	  strcpy(command,"gomo");
	  if (argc>0){
		  for (i=1;i<argc;++i){
			  if (lc+strlen(argv[i]+2)>=alloc_size){
				  alloc_size += strlen(argv[i])+2;
				  Resize(command, alloc_size, char);
			  }
			  command[lc++] = ' ';
			  for (j=0;j<strlen(argv[i]);++j)
				  command[lc++] = argv[i][j];
		  }
	  }
	  command[lc++] = '\0';
	  fprintf(xml_output,"<model>\n");
	  fprintf(xml_output,"<command_line>%s</command_line>\n",command);
	  fprintf(xml_output,"<significance_threshold>%g</significance_threshold>\n",e_threshold);
	  fprintf(xml_output,"</model>\n");
  }
  myfree(command);
}

/*************************************************************************
 * write_combined_organism_output()
 *
 * write the output for the combined p_values of several organisms
 * submitted
 *************************************************************************/
void write_combined_organism_output(
		FILE* text_output,			/* the text file to write the results in */
		FILE* xml_output,			/* the xml file to write the results in */
		HASH_TABLE motif_ht, 		/* the motif hashtable */
		HASH_TABLE godesc_ht, 		/* the go description hashtable */
		STRING_LIST_T* motifs, 		/* the list of motifs */
		STRING_LIST_T* all_go_terms, /* the list of go terms */
		int num_organisms,			/* the number of organisms */
		double e_threshold,			/* the  E-value threshold */
		BOOLEAN_T text_only
){
  int c,d,i,j,k,dummy;			/* bunch of counter variables */
  STRING_LIST_T* combined_set; 	/* keeps track of the go terms that exist in all organisms */
  HASH_TABLE_ENTRY* he;
  GOMO_RESULT_T* res = NULL;
  float pvalue;
  char *accession;
  
  fprintf(text_output,"# organism combined\n");
  if (!text_only){
	  fprintf(xml_output,"<species id=\"0\" name=\"combined\">\n");
	  fprintf(xml_output,"<input-files>\n<scoring-file>none</scoring-file>\n");
	  fprintf(xml_output,"<godb-file>none</godb-file>\n</input-files>\n");
  }
  // consider each motif
  for (i=0;i<get_num_strings(motifs);++i){
	  accession = get_nth_string(i,motifs);
	  he = hash_lookup_str(accession, motif_ht, &dummy);
	  if (he != NULL) {

		  if (!text_only){
			  fprintf(xml_output,"<pattern accession=\"%s\" name=\"%s\">\n",accession,accession);
		  }
		  HASH_TABLE goterms = (HASH_TABLE)hash_get_entry_value(he);
		  combined_set = new_string_list();
		  /* determine the go terms that have been found in all organisms */
		  for (j=0;j<get_num_strings(all_go_terms);++j){
			  he = hash_lookup_str(get_nth_string(j,all_go_terms), goterms, &dummy);
			  if (he != NULL) {
				  c = 0;
				  double *pvalues = (double*)hash_get_entry_value(he);
				  for (k=0;k<num_organisms;++k){
					  if (pvalues[k]>0)
						  ++c;
				  }
				  if (c==num_organisms)
					  add_string(get_nth_string(j,all_go_terms),combined_set);
			  }
		  }

		  /* combine p_values of go terms that exist in all organisms */
		  d = get_num_strings(combined_set); // number of combined go terms
		  if (d>0)
			  Resize(res,d,GOMO_RESULT_T);

		  for (j=0;j<d;++j){
			  char* go_term = get_nth_string(j,combined_set);
			  pvalue = 0.0;
			  he = hash_lookup_str(get_nth_string(j,combined_set), goterms, &dummy);
			  if (he != NULL) {
				  c = 0;
				  double *pvalues = (double*)hash_get_entry_value(he);
				  for (k=0;k<num_organisms;++k){
					  pvalue *= pvalues[k];
				  }
			  }
			  res[j].score = pvalue;
			  res[j].goterm = go_term;
		  }
		  qsort(res,d,sizeof(GOMO_RESULT_T),gomo_result_cmp);

		  /* output go-terms*/
		  for (j=0;j<d;++j){
			  char* go_desc = "";
			  he = hash_lookup_str(res[j].goterm,godesc_ht,&dummy);
			  if (he != NULL && hash_get_entry_value(he) != NULL){
				  go_desc = (char*)(hash_get_entry_value(he));
			  }

			  // multiple hypothesis testing
			  if (e_threshold<=0.0 || res[j].score*d<e_threshold) {
				  fprintf(text_output,"%s\t%s\t%.3e\t%s\n",accession,res[j].goterm,res[j].score,go_desc);
				  if (!text_only){
					  fprintf(xml_output,"<goterm id=\"%s\" E-value=\"%.3e\" qvalue=\"%.3e\" description=\"%s\"/>\n",
						  res[j].goterm,res[j].score*d,res[j].score,go_desc);
				  }
			  }
		  }

		  if (!text_only){
			  fprintf(xml_output,"</pattern>\n");
		  }

		  // free go set
		  free_string_list(combined_set);
		  myfree(res);
	  }
  }
  if (!text_only){
	  fprintf(xml_output,"</species>\n");
	  fflush(xml_output);
  }
  fflush(text_output);
}


/*************************************************************************
 * Entry point for gomo
 *************************************************************************/
int main(int argc, char *argv[]) {

  STRING_LIST_T* selected_motifs = NULL;

  char *output_dirname = default_output_dirname;
  const char *TEXT_FILENAME = "gomo.txt";
  const char *XML_FILENAME = "gomo.xml";
  const char *HTML_STYLESHEET = "gomo.xsl";
  const char *HTML_FILENAME = "gomo.html";
  char *xml_path = NULL;

  char* go_term = NULL;
  FILE *text_output = NULL;       	/* destination for text output */
  FILE *xml_output = NULL;       	/* destination for xml output */
  BOOLEAN_T text_only = FALSE;
  BOOLEAN_T status = TRUE;
  HASH_TABLE_ENTRY *he = NULL;
  float e_threshold = SHOW_EVALUE;

  HASH_TABLE motif_ht = NULL;	// A hash_table recording gomo results for each motif
  STRING_LIST_T* motifs = new_string_list(); // all motifs used as keys in motif_ht

  clock_t c0, c1; /* measuring cpu_time */

  /**********************************************
   * COMMAND LINE PROCESSING
   **********************************************/

  const int num_options = 7;
  cmdoption const motif_scan_options[] = {
	{"motif",REQUIRED_VALUE},
	{"nostatus", NO_VALUE},
	{"o", REQUIRED_VALUE},
	{"oc", REQUIRED_VALUE},
	{"text", NO_VALUE},
	{"t", REQUIRED_VALUE},
	{"verbosity", REQUIRED_VALUE}
  };

  int option_index = 0;

  // Define the usage message.
  char usage[] =
    "USAGE: gomo [options] <scoring file> <go-term file> [<scoring file> <go-term file>]*\n"
    "\n"
    "   Options:\n"
	"     --motif <id> (default=all)\n"
	"     --nostatus FLAG prevent output of the progress report to the terminal\n"
	"     --o <output dir> name of the directory for output, will not replacing existing dir\n"
	"     --oc <output dir> name of the directory for output, will replace existing dir\n"
	"     --text FLAG output in text format only\n"
	"     --t <float> E-value threshold considered significant (default = 10, t <= 0.0 shows all results)\n"
	"     --verbosity [1|2|3|4] (default 2)\n"
	"\n";

  // Parse the command line.
  if (simple_setopt(argc, argv, num_options, motif_scan_options) != NO_ERROR) {
    die("Error processing command line options: option name too long.\n");
  }

  while (TRUE) {
    int c = 0;
    char* option_name = NULL;
    char* option_value = NULL;
    const char * message = NULL;

    // Read the next option, and break if we're done.
    c = simple_getopt(&option_name, &option_value, &option_index);
    if (c == 0) {
      break;
    }
    else if (c < 0) {
      (void) simple_getopterror(&message);
      die("Error processing command line options (%s)\n", message);
    }
    if (strcmp(option_name, "verbosity") == 0){
      verbosity = atoi(option_value);
    }
    else if (strcmp(option_name, "motif") == 0){
      if (selected_motifs == NULL) {
        selected_motifs = new_string_list();
      }
      add_string(option_value, selected_motifs);
    }
    else if (strcmp(option_name, "o") == 0){
    	output_dirname = option_value;
    }
    else if (strcmp(option_name, "t") == 0){
		float x = atof(option_value);
		e_threshold = x;
    }
    else if (strcmp(option_name, "oc") == 0){
    	output_dirname = option_value;
    	clobber = TRUE;
    }
    else if (strcmp(option_name, "text") == 0){
    	text_only = TRUE;
	}
    else if (strcmp(option_name, "nostatus") == 0){
    	status = FALSE;
    }
  }

  // Must have scoring file and go term file
  if (argc < option_index + 2 || (argc - option_index) % 2 != 0) {
    fprintf(stderr, usage);
    exit(EXIT_FAILURE);
  }

  STRING_LIST_T* scoring_filenames = new_string_list();
  STRING_LIST_T* goterm_filenames = new_string_list();
  while (option_index < argc){
	  add_string(argv[option_index],scoring_filenames);
	  option_index++;
	  add_string(argv[option_index],goterm_filenames);
	  option_index++;
  }

  set_alphabet(verbosity, "ACGT");

  // measure time
  c0 = clock();


  /**********************************************
   * Get an idea about the number of go terms to
   * handle
   **********************************************/
  STRING_LIST_T* all_go_terms = new_string_list();
  get_unique_go_terms(goterm_filenames, all_go_terms);
  // A hash_table mapping of go terms to its description
  HASH_TABLE godesc_ht = hash_create(get_num_strings(all_go_terms)*2);

  /**********************************************
   * prepare output
   **********************************************/
  if (text_only) {
    // Legacy: plain text output to standard out.
    text_output = stdout;
  }
  else {
	  // allow clobbering of the default output directory
	  if (strcmp(output_dirname,default_output_dirname)==0) {
	 	  clobber = TRUE;
	   }
	   if (create_output_directory(output_dirname, clobber, status)) {
	 	  // Failed to create output directory.
	 	  exit(1);
	   }
	   // Create the name of the text output file
	   // "<dir>/text_filename/" and open it for writing
	   char *txt_path = make_path_to_file(output_dirname, TEXT_FILENAME);
	   text_output = fopen(txt_path, "w"); //FIXME CEG check for errors
	   xml_path = make_path_to_file(output_dirname, XML_FILENAME);
	   xml_output = fopen(xml_path, "w");
	   myfree(txt_path);
  }

  /**********************************************
   * Start output
   **********************************************/
  start_output(text_output, xml_output, text_only, program_name, e_threshold, argc, argv);


  /**********************************************
   * Run gomo on each dataset
   **********************************************/
  int org_index,i,j,k,dummy;
  CISML_T *cisml;

  // process each organism
  int num_organisms = get_num_strings(scoring_filenames);
  for (org_index=0; org_index<num_organisms;++org_index){
	  // a key list for the hash containing all the go-terms for one organism
	  STRING_LIST_T* organism_go_terms = new_string_list();

	  /**********************************************
	   * Read the scored sequence list.
	   **********************************************/

	  if (verbosity >= HIGH_VERBOSE){
		  fprintf(stderr, "Reading in %s \n", get_nth_string(org_index, scoring_filenames));
	  }
	  cisml = read_cisml(get_nth_string(org_index,scoring_filenames));
	  if (verbosity >= NORMAL_VERBOSE){
		  fprintf(stderr, "Sequence file %s \n", get_cisml_program_name(cisml));
	  }

	  if (cisml == NULL) {
		// Wasn't cisml XML
		die("The scoring file needs to be in CisML format: %s", get_nth_string(org_index,scoring_filenames));
	  }

	  // estimate number of motifs and create hashtable
	  if (num_organisms >= 1 && org_index==0){
		  motif_ht = hash_create(get_cisml_num_patterns(cisml)*2);
	  }

	  /**********************************************
	   * Read the go term mapping.
	   **********************************************/
	  HASH_TABLE go_ht = hash_create(get_num_strings(all_go_terms)*2); 	// A hash_table recording go terms
	  read_go_file(get_nth_string(org_index,goterm_filenames), organism_go_terms, &go_ht, &godesc_ht);

	  /**********************************************
	   * Perform gomo calculations
	   **********************************************/

	  fprintf(text_output,"# organism %d\n",org_index);
	  // add organism to output
	  if (!text_only) {
		  fprintf(xml_output,"<species id=\"%d\" name=\"%d\">\n",org_index+1, org_index+1);
		  fprintf(xml_output,"<input-files>\n<scoring-file>%s</scoring-file>\n",get_nth_string(org_index,scoring_filenames));
		  fprintf(xml_output,"<godb-file>%s</godb-file>\n</input-files>\n",get_nth_string(org_index,goterm_filenames));
	  }

	  // process each motif
	  for (i=0; i<get_cisml_num_patterns(cisml);++i){

		  char* accession = get_pattern_accession(get_cisml_patterns(cisml)[i]);
		  // check if only a particular motif should be scanned
		  if (selected_motifs != NULL &&  !have_string(accession,selected_motifs))
			  continue;

		  run_gomo_on_organism(
	  		cisml,
	  		get_cisml_patterns(cisml)[i],
	  		go_ht,
	  		godesc_ht,
	  		organism_go_terms,
	  		all_go_terms,
	  		e_threshold,
	  		org_index,
	  		num_organisms,
	  		motif_ht,
	  		status,
	  		text_only,
	  		xml_output,
	  		text_output,
	  		motifs
		  );
	  }

	  if (status){
	  	  fprintf(stderr, "Process organism %d: 100 %%\n",org_index+1);
	  	  fprintf(stderr, "Writing output\n");
	  	}

	  	if (!text_only){
	  	  fprintf(xml_output,"</species>\n");
	  	}

	  	// clean
	  	for (j=0;j<get_num_strings(organism_go_terms);++j){
	  	  go_term = get_nth_string(j,organism_go_terms);
	  	  HASH_TABLE_ENTRY *he = hash_lookup_str(go_term,go_ht,&dummy);
	  	  if (he != NULL && hash_get_entry_value(he) != NULL){
	  		  char* v = (char*)(hash_get_entry_value(he));
	  		  myfree(v);
	  	  }
	  	}
	  	hash_destroy(go_ht);
	  	free_cisml(cisml);
	  	free_string_list(organism_go_terms);
  }

  // write combined organism results
  if (num_organisms > 1)
	  write_combined_organism_output(text_output, xml_output, motif_ht, godesc_ht,
			  motifs, all_go_terms, num_organisms, e_threshold,text_only);
  fflush(text_output);

  // finish output
  if (!text_only){
	  fprintf(xml_output,"</gomo>\n");
	  fclose(xml_output);

	  /* html files are to */
	  if (e_threshold>0.0){
		  char *stylesheet_path = make_path_to_file(ETC_DIR, HTML_STYLESHEET);
		  char *html_path = make_path_to_file(output_dirname, HTML_FILENAME);
		  print_xml_filename_to_filename_using_stylesheet(
			    xml_path,        	/* path to XML input file IN */
			    stylesheet_path,	/* path to MEME XSL stylesheet IN */
			    html_path   		/* path to HTML output file IN */
		  );

		  myfree(html_path);
		  myfree(stylesheet_path);
	  } else {
		  fprintf(stderr,"Threshold value does not allow HTML output due to size and memory constraint.");
	  }

	  myfree(xml_path);
  }

  fclose(text_output);

  // measure time
  c1 = clock();
  if (verbosity >= NORMAL_VERBOSE) { // starting time
 	  fprintf(stderr,"cycles (CPU);            %ld cycles\n", (long) c1);
 	  fprintf(stderr,"elapsed CPU time:        %f seconds\n", (float) (c1 - c0)/CLOCKS_PER_SEC);
  }

  /**************************************
   * Clean up.
   **************************************/
  // free hashtables

  for (j=0;j<get_num_strings(all_go_terms);++j){
	  char* go_term = get_nth_string(j,all_go_terms);
	  HASH_TABLE_ENTRY *he = hash_lookup_str(go_term,godesc_ht,&dummy);
	  if (he != NULL && hash_get_entry_value(he) != NULL){
		  char* v = (char*)(hash_get_entry_value(he));
		  myfree(v);
	  }
  }
  hash_destroy(godesc_ht);

  //FIXME: free hash appropriately
  hash_destroy(motif_ht);

  free_string_list(all_go_terms);
  free_string_list(motifs);
  free_string_list(scoring_filenames);
  free_string_list(goterm_filenames);
  return(0);
}
